/*********************************************************************
 * Filename:      snmpd-mux.c
 * Version:       1.0.1 Final Release
 * Description:   Program to multiplex port 161 between snmp agents
 * Author:        Sveinar Rasmussen (sveinar@rasmussen.org)
 * Date:          Thu Aug 14 11:08:02 METDST 1997
 * Thanks to:     G.J. Drekker (for tips'n'tricks)
 * Notes:         Very quick and dirty program with no comments :)
 ********************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <signal.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

/*==================================================================+
** Defines, global variables and types
**=================================================================*/
#define	MAIN_SNMPPORT		161
#define CLIENT_TIMEOUT		10L
#define	MAXBUFLEN		2000
struct	client_str {
	char			*oid;
	int			fd;
	struct sockaddr_in	src_address;
	struct sockaddr_in	dst_address;
} clients[2];
typedef	struct client_str client;
struct sockaddr_in        main_address;
char			  buf[MAXBUFLEN];
extern int		  errno;
int			  mainfd, mainport = MAIN_SNMPPORT, numclients;
char                      extraAgentPort[20], defaultAgentPort[20];

/*==================================================================+
** Small error message functions
**=================================================================*/
void fatal(char *s) { perror(s); exit(1); }
void sys_error(char *s) { perror(s); }
void error(char *s, int a1) { fprintf(stderr, s, a1); fflush(stderr); }

/*==================================================================+
** Get Config Vars (port settings)
**=================================================================*/
void get_cfg_ports()
{
  FILE *cfg;
  char line[300];
  int i, j;

  extraAgentPort[0] = defaultAgentPort[0] = 0;
  if (!(cfg = fopen("../etc/snmpd.conf", "r")))
    if (!(cfg = fopen("etc/snmpd.conf", "r")))
      if (!(cfg = fopen("snmpd.conf", "r"))) {
	printf ("snmpd.conf not found in ../etc; etc; ./  -> Abort!\n");
	exit (-1);
      }  
  while (!feof(cfg)) {
    fgets ((char*)&line, 300, cfg);
    if (line[0]!='#') {
      if (strncmp(line, "ThisAgentSnmpdPort", 18)==0) {
	for (i=18; (line[i]<'0') || (line[i]>'9'); i++) ;
	for (j=i; (line[j]<='9') && (line[j]>='0'); j++) ;
	line[j] = 0;
	strcpy(extraAgentPort, &line[i]);
      }
      if (strncmp(line, "OtherAgentSnmpdPort", 19)==0) {
	for (i=19; (line[i]<'0') || (line[i]>'9'); i++) ;
	for (j=i; (line[j]<='9') && (line[j]>='0'); j++) ;
	line[j] = 0;
	strcpy(defaultAgentPort, &line[i]);
      }
    }
  }
  fclose(cfg);
  if (extraAgentPort[0]==0) {
    printf ("Error in config file: 'ThisAgentSnmpdPort' not set.\n");
    exit(-1);
  }
  if (defaultAgentPort[0]==0) {
    printf ("Error in config file: 'OtherAgentSnmpdPort' not set.\n");
    exit(-1);
  }
}

/*==================================================================+
** Open socket to given client (the extra agent + the main agent)
**=================================================================*/
int setclientsock(client *c, char *host, char *name)
{
  struct sockaddr_in src_addr;
  int	 option;
  int	 dst_port;
  struct servent *se;
  u_long addr;
  struct hostent *hp;
  char	 **p;
  
  c->src_address.sin_family = AF_INET;
  c->src_address.sin_port = htons(0);
  if (host == NULL) c->src_address.sin_addr.s_addr = htonl(INADDR_ANY);
  else {
    if ((addr = inet_addr(host)) == -1) {
      hp = gethostbyname(host);
      if (hp == NULL) exit (3);
    } else hp = gethostbyaddr((char *)&addr, sizeof (addr), AF_INET);

    if (hp == NULL) memcpy((char *)&c->src_address.sin_addr.s_addr,
			   (char *)&addr,
			   sizeof (c->src_address.sin_addr.s_addr));
    else {
      p = hp->h_addr_list;
      memcpy(&c->src_address.sin_addr.s_addr, *p,
	     sizeof (c->src_address.sin_addr.s_addr));
      p++;
      if (*p != NULL) printf("Alternative addresses for host %s ignored\n", host);
    }
  }
  
  if ((se = getservbyname(name, "udp"))) dst_port = ntohs(se->s_port);
  else {
    dst_port = atoi(name);
    if (dst_port == 0) {
      fprintf(stderr, "Illegal port: %s\n", name);
      exit(1);
    }
  }
  
  c->fd = socket(AF_INET, SOCK_DGRAM, 0);
  if (c->fd == -1) { fatal("Can not create socket"); return 0; }
  option = 1;
  if (setsockopt(c->fd, SOL_SOCKET, SO_REUSEADDR, (void *)&option,
		 sizeof(option)) < 0) 
    fatal("setsockopt client");
  
  src_addr.sin_family = AF_INET;
  src_addr.sin_port = htons(0);
  src_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  
  if (bind(c->fd, (struct sockaddr *)&src_addr,
	   sizeof(struct sockaddr_in)) < 0) {
    fatal("Can not bind client socket"); return 0; }
  
  c->dst_address.sin_family = AF_INET;
  c->dst_address.sin_port = htons(dst_port);
  c->dst_address.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
  return 1;
}

/*==================================================================+
** If stuff happens on sockets, receive data
**=================================================================*/
void discardclients(fd_set *fdsp) {
  int	             i;
  int	             recv_size;
  struct sockaddr_in socket;
  int		     socksize;
  
  for (i = 0; i < numclients; i++) 
    if (FD_ISSET(clients[i].fd, fdsp)) {
      socksize = sizeof(socket);
      recv_size = recvfrom(clients[i].fd, buf, MAXBUFLEN, 0,
			   (struct sockaddr *) &socket, &socksize);
      if (recv_size <= 0) {
	sys_error("Client discard");
      }
    }
}

/*==================================================================+
** Send stuff to originating client
**=================================================================*/
int sendclient(client *c, char *bufp, int size) {
  int		     send_size, recv_size;
  struct sockaddr_in socket;
  int		     socksize;
  fd_set             fds;
  int                fdcnt;
  struct timeval     tm;
  
  send_size = sendto(c->fd, bufp, size, 0, 
		     (struct sockaddr *) &c->dst_address,
		     sizeof(struct sockaddr_in));
  
  if (send_size < 0) sys_error("send failed");
  FD_ZERO(&fds);
  FD_SET(c->fd, &fds);
  
  tm.tv_usec = 0;
  tm.tv_sec = CLIENT_TIMEOUT;  
  fdcnt = select(FD_SETSIZE, &fds, 0 , 0, &tm);
  if (fdcnt < 0) sys_error("client select failed");
  if (fdcnt == 0) error("client select timed out\n", fdcnt);
  if (fdcnt > 0) {
    if (FD_ISSET(c->fd, &fds)) {
      socksize = sizeof(socket);
      recv_size = recvfrom(c->fd, buf, MAXBUFLEN, 0,
			   (struct sockaddr *) &socket, &socksize);
      if (recv_size <= 0) sys_error("Main receive failed");
      return recv_size;
    }
  }
  return 0;
}

/*==================================================================+
** Main procedure
**=================================================================*/
int main(int argc, char *argv[])
{
  int	         option, fdcnt, send_size, recv_size, socksize, i, last_48=0, almost_last = 0;
  fd_set         fds;
  struct timeval tm;
  struct	 sockaddr_in socketC;
  struct	 client_str *client;

  get_cfg_ports();
  setclientsock(&clients[0], "127.0.0.1", extraAgentPort);
  setclientsock(&clients[1], NULL, defaultAgentPort);
  numclients = 2;

  mainfd = socket(AF_INET, SOCK_DGRAM, 0);
  if (mainfd == -1) { fatal("Can not create mainsocket"); return 0; }
  option = 1;
  if (setsockopt(mainfd, SOL_SOCKET, SO_REUSEADDR, (void *)&option,
		 sizeof(option)) < 0) 
    fatal("setsockopt client");
  
  main_address.sin_family = AF_INET;
  main_address.sin_port = htons( MAIN_SNMPPORT );
  main_address.sin_addr.s_addr = INADDR_ANY;
  
  if (bind(mainfd, (struct sockaddr *)&main_address,
	   sizeof(struct sockaddr_in)) < 0) {
    sys_error("Can not bind main socket");
    fprintf(stderr, "Probably need to run as root to open port %d!\n", mainport);
    return 0;
  }

  while(1) {
    FD_ZERO(&fds);
    FD_SET(mainfd, &fds);
    for (i = 0; i < numclients; i++) FD_SET(clients[i].fd, &fds);
    
    tm.tv_usec = 0;
    tm.tv_sec = 10L;    
    fdcnt = select(FD_SETSIZE, &fds, 0 , 0, NULL);
    if (fdcnt < 0) sys_error("Select failed");
    discardclients(&fds);
    
    if (fdcnt > 0) {
      if (FD_ISSET(mainfd, &fds)) {
	socksize = sizeof(socketC);
	recv_size = recvfrom(mainfd, buf, MAXBUFLEN, 0,
			     (struct sockaddr *) &socketC, &socksize);
	if (recv_size == 0) {
	  error("Main receive timed out\n", 0);
	  continue;
	}
	if (recv_size < 0) {
	  sys_error("Main receive failed");
	  continue;
	}

	for (i=0; i < recv_size; i++) if (buf[i]==48) { 
	  almost_last = last_48;
	  last_48 = i;
	}
	if ((recv_size - last_48) < 9) last_48 = almost_last;
	/* [1.3].6.1.2.1.25 = MIB2 host resources tree */
	if ( (buf[last_48+5]==6) && (buf[last_48+6]==1) && (buf[last_48+7]==2) 
	     && (buf[last_48+8]==1) && (buf[last_48+9]==25) ) client = &clients[0];
	else client = &clients[1]; 

	recv_size = sendclient(client, buf, recv_size);
	if (recv_size == 0) {
	  error("Client receive timed out\n", 0);
	  continue;
	}
	if (recv_size <= 0) {
	  sys_error("Client receive failed");
	  continue;
	}

	send_size = sendto(mainfd, buf, recv_size, 0, 
			   (struct sockaddr *) &socketC, sizeof(socketC));
	if (send_size != recv_size) sys_error("Send error");
      }
    }
  }
  exit (-1);
}
